/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.formsolutions;

import static de.ozgcloud.eingang.common.formdata.FormSolutionsTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.SpyBean;
import org.springframework.test.context.ActiveProfiles;

import de.itvsh.kop.common.binaryfile.TempFileUtils;
import de.itvsh.kop.common.test.TestUtils;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.semantik.enginebased.FilesMapperHelper;
import de.ozgcloud.eingang.semantik.enginebased.formsolutions.FormSolutionsEngineBasedAdapter;
import lombok.SneakyThrows;

@SpringBootTest
@ActiveProfiles({ "local", "itcase" })
class FormSolutionsRequestMapperITCase {

	private static final String COMPONENTS = "components";
	private static final String PANELS = "panels";

	@SpyBean
	private FormSolutionsRequestMapper mapper;

	@Nested
	class TestParseExampleData {
		@Test
		void shouldParseData() {
			var parsed = parseRequestData(SIMPLE_JSON_DATA);

			assertThat(parsed).isNotNull();
		}

		@Test
		void shouldContainRawData() {
			var parsed = parseRequestData(SIMPLE_JSON_DATA);

			assertThat(parsed.getFormData()).isNotNull();
		}

		@Nested
		class TestWithPanels {
			@Test
			void shouldContainPanel() {
				var parsed = parseAndGetPanel();

				assertThat(parsed).isNotNull();
			}

			@Test
			void shouldHaveIdentifier() {
				var panel = parseAndGetPanel();

				assertThat(panel.get(FormSolutionsEngineBasedAdapter.IDENTIFIER)).isNotNull();
			}

			@Test
			@SuppressWarnings("unchecked")
			void shouldHaveComponents() {
				var panel = parseAndGetPanel();
				var components = (List<Map<String, Object>>) panel.get(COMPONENTS);

				assertThat(components).hasSize(2);
			}
		}

		private Map<String, Object> parseAndGetPanel() {
			return FormSolutionsRequestMapperITCase.this.parseAndGetPanel(SIMPLE_JSON_DATA);
		}
	}

	@Nested
	class TestParseNestedComponents {

		@Test
		void shouldParseData() {
			var parsed = parseRequestData(NESTED_COMPONENTS_JSON);

			assertThat(parsed).isNotNull();
		}

		@Test
		@SuppressWarnings("unchecked")
		void shouldHaveComponents() {
			var panel = parseAndGetPanel();

			assertThat((List<Map<String, Object>>) panel.get(COMPONENTS)).hasSize(1);
		}

		@Test
		void shouldHaveIdentifier() {
			var component = parseAndGetComponent();

			assertThat(component).containsEntry(FormSolutionsEngineBasedAdapter.IDENTIFIER, OBJEKTGRUPPE_0);
		}

		@Test
		void shouldHaveNestedComponents() {
			var component = parseAndGetComponent();

			assertThat(component.keySet()).hasSize(3);
		}

		@SuppressWarnings("unchecked")
		private Map<String, Object> parseAndGetComponent() {
			return ((List<Map<String, Object>>) parseAndGetPanel().get(COMPONENTS)).get(0);
		}

		private Map<String, Object> parseAndGetPanel() {
			return FormSolutionsRequestMapperITCase.this.parseAndGetPanel(NESTED_COMPONENTS_JSON);
		}

	}

	@Nested
	class TestParsePdfRepresentation {

		private static final String PARSED_REPRESENTATIONS = "parsedRepresentations";
		private static final byte[] DECODED_VALUE = "Lorem ipsum".getBytes();

		@Test
		@SneakyThrows
		void shouldHavePdf() {
			var parsed = parseRequestData(PDF_REPRESENTATION_JSON);

			assertThat(getRepresentation(parsed.getFormData())).isNotNull();

			assertThat(TestUtils.contentStreamToByteArray(getRepresentation(parsed.getFormData()).getContentStream()))
					.isEqualTo(DECODED_VALUE);
		}

		@SuppressWarnings("unchecked")
		private IncomingFile getRepresentation(Map<String, Object> data) {
			return ((List<IncomingFile>) ((Map<String, Object>) data.get(FilesMapperHelper.FIELD_NAME_MAPPED_FILES)).get(PARSED_REPRESENTATIONS))
					.get(0);
		}
	}

	@Nested
	class TestParseAttachmentZip {

		private static final String PARSED_ATTACHMENTS = "parsedAttachments";
		private static final byte[] DECODED_VALUE = "Lorem ipsum".getBytes();

		@Test
		@SneakyThrows
		void shouldHaveZip() {
			var parsed = parseRequestData(ZIP_ATTACHMENT_JSON);

			assertThat(getAttachment(getFiles(parsed.getFormData()))).isNotNull();

			assertThat(TestUtils.contentStreamToByteArray(getAttachment(getFiles(parsed.getFormData())).getContentStream()))
					.isEqualTo(DECODED_VALUE);
		}

		@SuppressWarnings("unchecked")
		private IncomingFile getAttachment(Map<String, Object> files) {
			return ((List<IncomingFileGroup>) files.get(PARSED_ATTACHMENTS)).get(0).getFiles().get(0);
		}
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> parseAndGetPanel(String json) {
		var data = (Map<String, Object>) parseRequestData(json).getFormData().get(FormSolutionsEngineBasedAdapter.ASSISTANT);
		return ((List<Map<String, Object>>) data.get(PANELS)).get(0);
	}

	@SuppressWarnings("unchecked")
	private Map<String, Object> getFiles(Map<String, Object> data) {
		return (Map<String, Object>) data.get(FilesMapperHelper.FIELD_NAME_MAPPED_FILES);
	}

	private FormData parseRequestData(String json) {
		var file = TempFileUtils.writeTmpFile(json);
		var result = mapper.map(file);
		file.delete();
		return result;
	}
}
