/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.formsolutions;

import static de.ozgcloud.eingang.common.formdata.FormSolutionsTestFactory.*;
import static de.ozgcloud.eingang.semantik.enginebased.formsolutions.FormSolutionsEngineBasedAdapter.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.File;
import java.io.InputStream;
import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.Spy;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.type.TypeReference;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;

import de.itvsh.kop.common.binaryfile.TempFileUtils;
import de.itvsh.kop.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.semantik.enginebased.FilesMapperHelper;
import lombok.SneakyThrows;

class FormSolutionsRequestMapperTest {

	private static final String TRANSACTION_ID = "transactionId";
	private static final String COMPONENTS = "components";
	private static final String STRING_VALUE = "stringValue";
	private static final String PANELS = "panels";
	private static final String ZUSTAENDIGE_STELLE = "zustaendigeStelle";
	private static final String POSTKORBHANDLE = "postkorbhandle";

	@Spy
	@InjectMocks
	private FormSolutionsRequestMapper mapper;

	@Mock
	private FormSolutionsAttachmentsMapper attachmentMapper;
	@Mock
	private FormSolutionsRepresentationsMapper representationsMapper;

	@Spy
	private ObjectMapper objectMapper = new ObjectMapper();

	private File simpleJsonFile;
	private File nestedComponenetJsonFile;

	@BeforeEach
	void writeJsonFile() {
		simpleJsonFile = TempFileUtils.writeTmpFile(SIMPLE_JSON_DATA);
		nestedComponenetJsonFile = TempFileUtils.writeTmpFile(NESTED_COMPONENTS_JSON);
	}

	@AfterEach
	void delJsonFile() {
		simpleJsonFile.delete();
		nestedComponenetJsonFile.delete();
	}

	@Nested
	class TestJsonToMapMapping {

		@Test
		void shouldMap() {
			var dataMap = mapper.mapFormData(simpleJsonFile);

			assertThat(dataMap).isNotNull();
		}

		@Test
		void shouldContainZustaendigeStelle() {
			var dataMap = mapper.mapFormData(simpleJsonFile);

			assertThat((String) dataMap.get(ZUSTAENDIGE_STELLE)).isEqualTo(ORGANISATIONSEINHEITEN_ID);
		}

		@Test
		void shouldContainsPostfachId() {
			var dataMap = mapper.mapFormData(simpleJsonFile);

			assertThat((String) dataMap.get(POSTKORBHANDLE)).isEqualTo(POSTFACH_ID);
		}

		@Test
		void shouldContainRequestId() {
			var dataMap = mapper.mapFormData(simpleJsonFile);

			assertThat((String) dataMap.get(TRANSACTION_ID)).isEqualTo(FORM_ID_VALUE);
		}

		@Test
		void shouldContainAssitant() {
			var dataMap = mapper.mapFormData(simpleJsonFile);

			assertThat(dataMap.get(ASSISTANT)).isNotNull();
		}

		@Test
		@SneakyThrows
		void shouldHandleJsonException() throws JsonMappingException, JsonProcessingException {
			doThrow(JsonProcessingException.class).when(objectMapper).readValue(any(InputStream.class),
					Mockito.<TypeReference<Map<String, Object>>>any());

			assertThatThrownBy(() -> mapper.mapFormData(simpleJsonFile)).isInstanceOf(TechnicalException.class);
		}

		@Test
		@SuppressWarnings("unchecked")
		void shouldContainFormIdentifier() {
			var dataMap = mapper.mapFormData(simpleJsonFile);

			assertThat((String) ((Map<String, Object>) dataMap.get(ASSISTANT))
					.get(IDENTIFIER))
							.isEqualTo(IDENTIFIER_VALUE);
		}

		@Nested
		class TestPanels {
			@Test
			void shouldContainPanels() {
				var dataMap = mapper.mapFormData(simpleJsonFile);

				assertThat(getPanels(dataMap)).isNotNull();
			}

			@Test
			void shouldContainPanelIdentifier() {
				var dataMap = mapper.mapFormData(simpleJsonFile);

				assertThat(getPanels(dataMap).get(0)).containsEntry(IDENTIFIER, PANEL_ID);
			}

			@Test
			void shouldContainPanelComponets() {
				var dataMap = mapper.mapFormData(simpleJsonFile);

				assertThat(getPanels(dataMap).get(0).get(COMPONENTS)).isNotNull();
			}

			@Test
			void shouldContainTextComponets() {
				var dataMap = mapper.mapFormData(simpleJsonFile);

				assertThat(getComponents(dataMap).get(0))
						.containsEntry(IDENTIFIER, COMPONENT_ID)
						.containsEntry(STRING_VALUE, COMPONENT_VALUE);
			}

			@Test
			void shouldContainDateComponets() {
				var dataMap = mapper.mapFormData(simpleJsonFile);

				assertThat(getComponents(dataMap).get(1))
						.containsEntry(IDENTIFIER, DATE_COMPONENT_ID)
						.containsEntry(STRING_VALUE, DATE_COMPONENT_VALUE);
			}

			@Nested
			class TestNestedPanels {
				@Test
				void shouldContainGroup() {
					var dataMap = mapper.mapFormData(nestedComponenetJsonFile);

					assertThat(getComponents(dataMap).get(0)).containsEntry(IDENTIFIER, OBJEKTGRUPPE_0);
				}

				@Test
				void shouldContainDateField() {
					var dataMap = mapper.mapFormData(nestedComponenetJsonFile);

					assertThat(getNestedComponents(dataMap).get(0))
							.containsEntry(IDENTIFIER, DATE_COMPONENT_ID)
							.containsEntry(STRING_VALUE, DATE_COMPONENT_VALUE);
				}
			}
		}

		@SuppressWarnings("unchecked")
		private List<Map<String, Object>> getComponents(Map<String, Object> dataMap) {
			return (List<Map<String, Object>>) getPanels(dataMap).get(0).get(COMPONENTS);
		}

		@SuppressWarnings("unchecked")
		private List<Map<String, Object>> getNestedComponents(Map<String, Object> dataMap) {
			return (List<Map<String, Object>>) ((List<Map<String, Object>>) getPanels(dataMap).get(0).get(COMPONENTS)).get(0).get(COMPONENTS);
		}

		@SuppressWarnings("unchecked")
		private List<Map<String, Object>> getPanels(Map<String, Object> dataMap) {
			return (List<Map<String, Object>>) ((Map<String, Object>) dataMap.get(ASSISTANT)).get(PANELS);
		}
	}

	@Nested
	class TestFileMapping {

		@Nested
		class TestMapFiles {

			private File jsonFile;

			@BeforeEach
			void writeJsonFile() {
				jsonFile = TempFileUtils.writeTmpFile(ATTACHMENTS_JSON);
			}

			@AfterEach
			void deleteTempFile() {
				jsonFile.delete();
			}

			@Test
			void shouldCallAttachmentMappers() {
				mapper.mapFiles(FormDataTestFactory.create(), jsonFile);

				verify(attachmentMapper).mapAttachments(Mockito.<Map<String, Object>>any());
			}

			@DisplayName("result should have mapped files field")
			@Test
			void shouldHaveMappedFilesField() {
				var result = mapper.mapFiles(FormDataTestFactory.create(), jsonFile);

				assertThat(result.getFormData()).containsKey(FilesMapperHelper.FIELD_NAME_MAPPED_FILES);
			}

			@Test
			void shouldCallRepresentationMapper() {
				mapper.mapFiles(FormDataTestFactory.create(), jsonFile);

				verify(representationsMapper).mapRepresentations(Mockito.<Map<String, Object>>any(), eq(jsonFile));
			}

			@Test
			void shouldRemoveZip() {
				var formData = FormDataTestFactory.withFormDataMaps(Map.of(FormSolutionsAttachmentsMapper.ZIP, "test"));

				var dataMap = mapper.mapFiles(formData, jsonFile);

				assertThat(dataMap.getFormData()).doesNotContainKey(FormSolutionsAttachmentsMapper.ZIP);
			}

			@Test
			void shouldRemovePdf() {
				var formData = FormDataTestFactory.withFormDataMaps(Map.of(FormSolutionsRepresentationsMapper.PDF, "test"));

				var dataMap = mapper.mapFiles(formData, jsonFile);

				assertThat(dataMap.getFormData()).doesNotContainKey(FormSolutionsRepresentationsMapper.PDF);
			}
		}

	}

}
