package de.ozgcloud.eingang.formsolutions;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.List;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.test.annotation.DirtiesContext;
import org.springframework.test.context.ActiveProfiles;
import org.springframework.test.util.ReflectionTestUtils;

import de.itvsh.kop.common.test.TestUtils;
import de.itvsh.ozg.pluto.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.itvsh.ozg.pluto.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.itvsh.ozg.pluto.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import de.itvsh.ozg.pluto.vorgang.GrpcCreateVorgangResponse;
import de.itvsh.ozg.pluto.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import de.ozgcloud.eingang.router.PlutoServerResolver;
import io.grpc.Channel;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.ClientResponseObserver;

@SpringBootTest
@DirtiesContext
@ActiveProfiles({ "local", "itcase" })
public class FormsolutionsITCase {

	@Autowired
	private SendFormEndpoint endpoint;

	@MockBean
	private PlutoServerResolver resolver;

	@Mock
	private Request request;

	@Mock
	private VorgangServiceBlockingStub blockingStub;
	@Mock
	private BinaryFileServiceStub fileStub;
	@Mock
	private CallStreamObserver<GrpcUploadBinaryFileRequest> fileStreamObserver;

	@Nested
	class TestReceiveFormWithAttachments {

		@Captor
		private ArgumentCaptor<ClientResponseObserver<GrpcUploadBinaryFileRequest, GrpcUploadBinaryFileResponse>> observerCaptor;

		@Captor
		private ArgumentCaptor<GrpcUploadBinaryFileRequest> requestCaptor;

		@BeforeEach
		void init() {
			when(request.getJSON()).thenReturn(loadTextFile("SimpleJsonWithAttachments.json"));
		}

		@BeforeEach
		void initVorgangManagerResolver() {
			when(resolver.resolveVorgangServiceBlockingStubByOrganisationseinheitenId(any())).thenReturn(blockingStub);
			when(resolver.resolveBinaryFileServiceStubByOrganisationsEinheitId(any())).thenReturn(fileStub);
			
			Channel mockChannel = mock(Channel.class);
			when(blockingStub.getChannel()).thenReturn(mockChannel );
			when(blockingStub.startCreation(any())).thenReturn(GrpcCreateVorgangResponse.newBuilder().setVorgangId("42").build());
	
			when(fileStub.uploadBinaryFileAsStream(any())).thenReturn(fileStreamObserver);
		}

		@Test
		void shouldSendContentOfAttachment() {
			new Thread(() -> endpoint.receiveForm(request)).start();

			var requests = getFileRequests();

			var fileContent = requests.get(1).getFileContent();
			assertThat(fileContent.isEmpty()).isFalse();
		}

		@Test
		void shouldHaveContentType() {
			new Thread(() -> endpoint.receiveForm(request)).start();

			var requests = getFileRequests();

			var contentType = requests.get(0).getMetadata().getContentType();
			assertThat(contentType).isEqualTo("application/pdf");
		}

		@Test
		void shouldHaveFileSize() {
			new Thread(() -> endpoint.receiveForm(request)).start();

			var requests = getFileRequests();

			var size = requests.get(0).getMetadata().getSize();
			assertThat(size).isEqualTo(6788);
		}

		private List<GrpcUploadBinaryFileRequest> getFileRequests() {
			verify(fileStub, timeout(1000)).uploadBinaryFileAsStream(observerCaptor.capture());
			var onReadyHandler = (Runnable) ReflectionTestUtils.getField(observerCaptor.getValue(), "onReadyHandler");
			onReadyHandler.run();

			verify(fileStreamObserver, times(2)).onNext(requestCaptor.capture());
			var requests = requestCaptor.getAllValues();
			return requests;
		}

	}

	private static String loadTextFile(final String fileName) {
		try {
			return IOUtils.toString(loadFile(fileName), StandardCharsets.UTF_8);
		} catch (IOException e) {
			throw new RuntimeException(e); // NOSONAR
		}
	}

	private static InputStream loadFile(final String fileName) {
		InputStream stream = TestUtils.class.getClassLoader().getResourceAsStream(fileName);
		if (stream == null) {
			throw new RuntimeException("File '" + fileName + "' not found."); // NOSONAR
		}
		return stream;
	}
}
