/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.intelliform;

import java.io.ByteArrayInputStream;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;

import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import org.w3c.dom.Document;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.semantik.enginebased.FilesMapperHelper;
import lombok.RequiredArgsConstructor;

//TODO Naming prüfen - er scheint mir nicht semantik zu mappen und befindet sich auch nicht im entsprechenden Modul
@Component
@RequiredArgsConstructor
class SemantikFormDataMapper {
	static final String FILE = "file";

	static final String CONTENT_FORMDATA_FILENAME_SUFFIX = "-Daten.xml";

	static final String ATTACHMENTS_FIELD = "depositAttachments";

	static final String ATTACHMENT_ID = "id";
	static final String ATTACHMENT_NAME = "name";
	static final String ATTACHMENT_DESCRIPTION = "description";
	static final String ATTACHMENT_ATTRIBUTES = "attributes";
	static final String ATTACHMENT_CONTENT = "content";
	static final String ATTACHMENT_CONTENT_TYPE = "contentType";

	static final String HEADER_FIELD = "header";

	static final String HEADER_ID = "t:id";
	static final String HEADER_TIMESTAMP = "t:timestamp";
	static final String HEADER_FORM_ID = "t:form-id";
	static final String HEADER_FORM = "t:form";
	static final String HEADER_SENDER = "t:sender";
	static final String HEADER_CUSTOMER = "t:customer";
	static final String HEADER_CUSTOMER_ID = "t:customer-id";
	static final String HEADER_CLIENT = "t:client";
	static final String HEADER_CLIENT_ID = "t:client-id";

	@Autowired
	private final XmlToJavaMapsMapper xmlToJavaMapsMapper;

	@Autowired
	private final FormDataIncomingFileMapper formDataIncomingFileMapper;

	@Autowired
	private final RepresentationsCalculator incomingFilesService;

	@Autowired
	private final DepositRequestIncomingFileMapper depositRequestFilesMapper;

	@Autowired
	private final AttachmentsContentAdder attachmentsContentAdder;

	public FormData mapToFormData(Deposit depositData) {
		var xmlFormDataStream = getXmlFormData(depositData.getData());

		var document = xmlToJavaMapsMapper.parseAsW3cDocument(new ByteArrayInputStream(xmlFormDataStream));
		var formDataMap = xmlToJavaMapsMapper.mapXmlToJavaMaps(document);

		addFiles(document, depositRequestFilesMapper.mapFiles(depositData), formDataMap);

		addHeader(document, formDataMap);

		return FormData.builder().formData(Collections.unmodifiableMap(formDataMap)).build();
	}

	private byte[] getXmlFormData(DepositData depositData) {
		var primaryId = depositData.getPrimaryDataAttachmentId();

		return depositData.getAttachments().stream()
				.filter(attachment -> StringUtils.equals(attachment.getId(), primaryId))
				.map(Attachment::getContent)
				.findFirst()
				.orElseThrow(() -> new RuntimeException("Request does not contain primary data attachment"));
	}

	private void addFiles(Document document, List<IncomingFile> depositRequestFiles, Map<String, Object> formDataMap) {
		List<IncomingFileGroup> attachments = formDataIncomingFileMapper.mapAttachments(document);
		attachments = attachmentsContentAdder.addContentToAttachments(attachments, depositRequestFiles);

		List<IncomingFile> representations = incomingFilesService.calculateRepresentations(attachments, depositRequestFiles);

		formDataMap.put(FilesMapperHelper.FIELD_NAME_MAPPED_FILES, Map.of(
				FilesMapperHelper.ATTACHMENTS, attachments,
				FilesMapperHelper.REPRESENTATIONS, representations));

		removeMappedFileReferences(attachments, formDataMap);
	}

	void removeMappedFileReferences(List<IncomingFileGroup> attachments, Map<String, Object> formDataMap) {
		attachments.forEach(group -> group.getFiles().stream().map(inFile -> getReferenceName(inFile.getVendorId(), formDataMap)).toList().stream()
				.forEach(ref -> ref.ifPresent(entry -> formDataMap.remove(entry.getKey()))));
	}

	@SuppressWarnings("unchecked")
	Optional<Map.Entry<String, Object>> getReferenceName(String vendorId, Map<String, Object> formDataMap) {
		return formDataMap.entrySet().stream().filter(entry -> {
			Object file = null;

			if (entry.getValue() instanceof Map) {
				file = ((Map<String, Object>) entry.getValue()).get(FILE);
			}

			return Objects.nonNull(file) && ((Map<String, String>) file).containsValue(vendorId);

		}).findFirst();
	}

	private void addHeader(Document document, Map<String, Object> formData) {
		formData.put(HEADER_FIELD, createHeaderMap(document));
	}

	private Map<String, Object> createHeaderMap(Document document) {
		var map = new HashMap<String, Object>();
		map.put(HEADER_ID, document.getDocumentElement().getAttribute(HEADER_ID));
		map.put(HEADER_TIMESTAMP, document.getDocumentElement().getAttribute(HEADER_TIMESTAMP));
		map.put(HEADER_FORM_ID, document.getDocumentElement().getAttribute(HEADER_FORM_ID));
		map.put(HEADER_FORM, document.getDocumentElement().getAttribute(HEADER_FORM));
		map.put(HEADER_SENDER, document.getDocumentElement().getAttribute(HEADER_SENDER));
		map.put(HEADER_CUSTOMER, document.getDocumentElement().getAttribute(HEADER_CUSTOMER));
		map.put(HEADER_CUSTOMER_ID, document.getDocumentElement().getAttribute(HEADER_CUSTOMER_ID));
		map.put(HEADER_CLIENT, document.getDocumentElement().getAttribute(HEADER_CLIENT));
		map.put(HEADER_CLIENT_ID, document.getDocumentElement().getAttribute(HEADER_CLIENT_ID));
		return map;
	}
}