/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.intelliform;

import static de.ozgcloud.eingang.intelliform.XmlDaten1Container.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.context.ApplicationContext;
import org.springframework.core.io.Resource;
import org.springframework.ws.test.server.MockWebServiceClient;
import org.springframework.ws.test.server.RequestCreators;
import org.springframework.ws.test.server.ResponseActions;
import org.springframework.ws.test.server.ResponseMatchers;

import de.itvsh.ozg.pluto.vorgang.GrpcEingang;
import de.itvsh.ozg.pluto.vorgang.GrpcFormData;
import de.itvsh.ozg.pluto.vorgang.GrpcIncomingFile;
import de.itvsh.ozg.pluto.vorgang.GrpcIncomingFileGroup;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.PostfachAddressTestFactory;
import de.ozgcloud.eingang.router.VorgangRemoteService;
import lombok.SneakyThrows;

@SpringBootTest
class FormDataEndpointITCase {

	private final static String TEST_FILE_PATH = "classpath:itcase/";

	private final static String REQUEST = "EinfachesFormularZweiAnhaengeSoapRequest.xml";
	private final static String RESPONSE = "EinfachesFormularZweiAnhaengeSoapResponse.xml";

	private static final String SOAP_REQUEST_OTHER_NAME = "XML-Daten-1-other_name_SoapRequest.xml";

	@Autowired
	private ApplicationContext applicationContext;
	@MockBean
	private VorgangRemoteService vorgangRemoteService;

	@Captor
	private ArgumentCaptor<FormData> formDataCaptor;
	@Captor
	private ArgumentCaptor<GrpcEingang> grpcEingangCaptor;
	@Captor
	private ArgumentCaptor<Optional<String>> organisationsEinheitIdCaptor;

	private MockWebServiceClient mockClient;

	@BeforeEach
	void initTest() {
		mockClient = MockWebServiceClient.createClient(applicationContext);
	}

	@DisplayName("Send antrag with attachments")
	@Nested
	class TestAntragWithAttachments {

		@Test
		void shouldSendRequest() {
			sendRequest();

			verify(vorgangRemoteService).createVorgang(any(FormData.class), any(GrpcEingang.class), any());
		}

		@Nested
		class checkAntragsteller {

			@Test
			void validateAntragstellerEmail() {
				sendRequest();

				assertThat(grpcEingangCaptor.getValue().getAntragsteller().getEmail()).isEqualTo("schneider@helgeschneider.local");
			}

			@Test
			void shouldMapAntragstellerPostfachId() {
				sendRequest();

				assertThat(grpcEingangCaptor.getValue().getAntragsteller().getPostfachId()).isEqualTo("nameIdAsOsiPostfachIdV1");
			}
		}

		@Nested
		class checkAttachments {

			@Test
			void validateOrganisationsEinheitId() {
				sendRequest();

				assertThat(organisationsEinheitIdCaptor.getValue()).isPresent().hasValue("10363455");
			}

			@Test
			void checkAttachmentsCount() {
				sendRequest();

				assertThat(grpcEingangCaptor.getValue().getNumberOfAttachments()).isEqualTo(2);
			}

			@Test
			void checkAttachmentGroupCount() {
				sendRequest();

				assertThat(grpcEingangCaptor.getValue().getAttachmentsCount()).isEqualTo(2);
			}

			@Test
			void checkAttachmentGroup1Count() {
				sendRequest();

				GrpcIncomingFileGroup group = grpcEingangCaptor.getValue().getAttachments(0);

				assertThat(group.getFilesCount()).isEqualTo(1);
			}

			@Test
			void checkAttachmentGroup1Files() {
				sendRequest();

				GrpcIncomingFile file = grpcEingangCaptor.getValue().getAttachments(0).getFiles(0);

				assertThat(file.getId()).isNotNull();
				assertThat(file.getVendorId()).isEqualTo("assistants.E0FBA361C191F8B723949467AE302BEA24E4745E");
				assertThat(file.getName()).isEqualTo("Helge1.jpg");
				assertThat(file.getContentType()).isEqualTo("image/jpeg");
				assertThat(file.getContent().size()).isZero();
			}

			@Test
			void checkForAttachmentFileContentStream() {
				sendRequest();

				var fileStream = formDataCaptor.getValue().getAttachments().get(0).getFiles().get(0).getContentStream();

				assertThat(fileStream).isNotNull();
			}

			@Test
			void checkAttachmentGroup2Count() {
				sendRequest();

				GrpcIncomingFileGroup group = grpcEingangCaptor.getValue().getAttachments(1);

				assertThat(group.getFilesCount()).isEqualTo(1);
			}

			@Test
			void checkAttachmentGroup2Files() {
				sendRequest();

				GrpcIncomingFile file = grpcEingangCaptor.getValue().getAttachments(1).getFiles(0);

				assertThat(file.getId()).isNotNull();
				assertThat(file.getVendorId()).isEqualTo("assistants.52D79E5B2118D1740045AB87151535DCAD24E9A7");
				assertThat(file.getName()).isEqualTo("Helgetext2.odt");
				assertThat(file.getContentType()).isEqualTo("application/vnd.oasis.opendocument.text");
				assertThat(file.getContent().size()).isZero();
			}
		}

		@Nested
		class validateRepresentations {
			@Test
			void checkRepresentationsCount() {
				sendRequest();

				assertThat(grpcEingangCaptor.getValue().getRepresentationsCount()).isEqualTo(1);
				assertThat(grpcEingangCaptor.getValue().getNumberOfRepresentations()).isEqualTo(1);
			}

			@Test
			void checkRepresentation1() {
				sendRequest();

				GrpcIncomingFile file = grpcEingangCaptor.getValue().getRepresentations(0);

				assertThat(file.getId()).isNotNull();
				assertThat(file.getVendorId()).isEqualTo("myForm-xml");
				assertThat(file.getName()).isEqualTo("XML-Daten.xml");
				assertThat(file.getContentType()).isEqualTo("text/xml");
				assertThat(file.getContent().size()).isZero();
			}

			@Test
			void checkForRepresentationFileContentStream() {
				sendRequest();

				var fileStream = formDataCaptor.getValue().getRepresentations().get(0).getContentStream();

				assertThat(fileStream).isNotNull();
			}
		}

		@DisplayName("service konto")
		@Nested
		class TestServiceKonto {

			@Test
			void shouldReturnMappedServiceKonto() {
				var eingang = sendRequest();

				assertThat(eingang.getHeader().getServiceKonto()).isNotNull();
				assertThat(eingang.getHeader().getServiceKonto().getType()).isEqualTo("OSI");
			}

			@Test
			void shoulContainsPostfachAddresses() {
				var eingang = sendRequest();

				assertThat(eingang.getHeader().getServiceKonto().getPostfachAddressesList()).hasSize(1);
			}

			@Test
			void shoulReturnMappedPostfachAddress() {
				var postfachAddress = sendRequest().getHeader().getServiceKonto().getPostfachAddressesList().get(0);

				assertThat(postfachAddress.getVersion()).isEqualTo(PostfachAddressTestFactory.VERSION);
				assertThat(postfachAddress.getIdentifier().getPropertyList()).hasSize(1);
				assertThat(postfachAddress.getIdentifier().getProperty(0).getName()).isEqualTo("postfachId");
				assertThat(postfachAddress.getIdentifier().getProperty(0).getValue(0)).isEqualTo("nameIdAsOsiPostfachIdV1");
				assertThat(postfachAddress.getType()).isEqualTo(1);
			}
		}

		@SneakyThrows
		private GrpcEingang sendRequest() {
			sendWebserviceRequest(REQUEST);

			return grpcEingangCaptor.getValue();
		}
	}

	@Nested
	class TestOtherNameForFormData {

		@Test
		void shouldSucceed() { // NOSONAR contains andExpect of mockClient
			sendWebserviceRequest(SOAP_REQUEST_OTHER_NAME).andExpect(ResponseMatchers.noFault());
		}

		@Test
		void shouldHaveOrgaId() {
			sendWebserviceRequest(SOAP_REQUEST_OTHER_NAME);

			verify(vorgangRemoteService).createVorgang(any(), grpcEingangCaptor.capture(), any());

			assertThat(grpcEingangCaptor.getValue().getZustaendigeStelle().getOrganisationseinheitenId()).isEqualTo("0815");
		}

	}

	@Nested
	class TestKeepFormDataOrder {

		@Test
		void shouldKeepEingangFieldsOrder() {
			var grpcFormData = requestFormData();

			assertThat(grpcFormData.getFieldList()).isEqualTo(XmlDaten1Container.EINGANG_FIELDS);
		}

		@Test
		void shouldKeepZustaendigestelleFieldsOrder() {
			var formFields = requestFormData().getForm(0).getFieldList();

			assertThat(formFields).isEqualTo(XmlDaten1Container.ZUSTAENDIGESTELLE_FIELDS);
		}

		@Test
		void shouldKeepEmpfangendestelleFieldsOrder() {
			var formFields = requestFormData().getForm(1).getFieldList();

			assertThat(formFields).isEqualTo(XmlDaten1Container.EMPFANGENDESTELLE_FIELDS);
		}

		@Test
		void shouldKeepErklaerungenFieldsOrder() {
			var formFields = requestFormData().getForm(2).getFieldList();

			assertThat(formFields).isEqualTo(XmlDaten1Container.ERKLAERUNGEN_FIELDS);
		}

		@Test
		void shouldKeepAnsprechpartnerFieldsOrder() {
			var formFields = requestFormData().getForm(3).getSubForm(0).getFieldList();

			assertThat(formFields).isEqualTo(XmlDaten1Container.ANSPRECHPARTNER_FIELDS);
		}

		@Test
		void shouldKeepAnschriftFieldsOrder() {
			var formFields = requestFormData().getForm(3).getSubForm(0).getSubForm(0).getFieldList();

			assertThat(formFields).isEqualTo(XmlDaten1Container.ANSCHRIFT_FIELDS);
		}

		@Test
		void shouldKeepKontaktFieldsOrder() {
			var formFields = requestFormData().getForm(3).getSubForm(0).getSubForm(1).getFieldList();

			assertThat(formFields).isEqualTo(XmlDaten1Container.KONTAKT_FIELDS);
		}

		@Test
		void shouldKeepVerwaltungsleistungFieldsOrder() {
			var formFields = requestFormData().getForm(3).getSubForm(1).getSubForm(0).getFieldList();

			assertThat(formFields).isEqualTo(XmlDaten1Container.VERWALTUNGSLEISTUNG_FIELDS);
		}

		@Test
		void shouldKeepAusgewaehlteZustaendigestelleFieldsOrder() {
			var formFields = requestFormData().getForm(3).getSubForm(1).getSubForm(0).getSubForm(0).getFieldList();

			assertThat(formFields).isEqualTo(XmlDaten1Container.AUSGEWAEHLTE_ZUSTAENDIGESTELLE_FIELDS);
		}

		private GrpcFormData requestFormData() {
			sendWebserviceRequest(REQUEST_XML_NAME);

			verify(vorgangRemoteService).createVorgang(any(), grpcEingangCaptor.capture(), any());
			return grpcEingangCaptor.getValue().getFormData();
		}
	}

	@SneakyThrows
	private ResponseActions sendWebserviceRequest(String requestFileName) {
		var response = mockClient.sendRequest(RequestCreators.withSoapEnvelope(getResource(requestFileName)))
				.andExpect(ResponseMatchers.noFault())
				.andExpect(ResponseMatchers.payload(getResource(RESPONSE)));

		verify(vorgangRemoteService).createVorgang(formDataCaptor.capture(), grpcEingangCaptor.capture(), organisationsEinheitIdCaptor.capture());

		return response;
	}

	private Resource getResource(String fileName) {
		return applicationContext.getResource(TEST_FILE_PATH + fileName);
	}
}
