/*
 * Copyright (C) 2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import java.io.InputStream;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.apache.commons.io.IOUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import com.google.protobuf.ByteString;

import de.itvsh.kop.common.binaryfile.GrpcFileUploadUtils;
import de.itvsh.kop.common.errorhandling.TechnicalException;
import de.itvsh.ozg.pluto.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.itvsh.ozg.pluto.grpc.binaryFile.GrpcUploadBinaryFileMetaData;
import de.itvsh.ozg.pluto.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.itvsh.ozg.pluto.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import de.itvsh.ozg.pluto.grpc.command.GrpcCallContext;
import de.itvsh.ozg.pluto.vorgang.GrpcCreateVorgangRequest;
import de.itvsh.ozg.pluto.vorgang.GrpcEingang;
import de.itvsh.ozg.pluto.vorgang.GrpcFinishCreationRequest;
import de.itvsh.ozg.pluto.vorgang.GrpcIncomingFile;
import de.itvsh.ozg.pluto.vorgang.GrpcIncomingFileGroup;
import de.itvsh.ozg.pluto.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@Service
public class VorgangRemoteService {

	@Autowired
	private PlutoServerResolver plutoServerResolver;

	public String createVorgang(FormData formData, GrpcEingang eingang, Optional<String> organisationsEinheitenId) {
		var plutoStub = plutoServerResolver.resolveVorgangServiceBlockingStubByOrganisationseinheitenId(organisationsEinheitenId);
		LOG.info("Connecting to pluto server " + plutoStub.getChannel().authority() + "; OrganisationsEinheitId: " + organisationsEinheitenId);

		return createVorgang(formData, eingang, plutoStub, getBinaryFileServiceStub(organisationsEinheitenId));
	}

	public String createVorgang(FormData formData, GrpcEingang eingang, VorgangServiceBlockingStub remoteStub,
			BinaryFileServiceStub remoteAsyncStub) {
		return new VorgangCreator(formData, eingang, remoteStub, remoteAsyncStub).create();
	}

	private BinaryFileServiceStub getBinaryFileServiceStub(Optional<String> organisationsEinheitenId) {
		return plutoServerResolver.resolveBinaryFileServiceStubByOrganisationsEinheitId(organisationsEinheitenId);
	}

	@RequiredArgsConstructor
	public class VorgangCreator {

		static final String VORGANG_ATTACHMENT_FIELD = "vorgangAttachment";
		static final String CALL_CONTEXT_CLIENT = "eingangAdpater";

		private final FormData formData;
		private final GrpcEingang eingang;
		private final VorgangServiceBlockingStub vorgangRemoteStub;
		private final BinaryFileServiceStub binaryFileRemoteStub;

		@Getter
		private String vorgangId;
		@Getter
		private List<IncomingFileGroup> uploadedAttachments;
		@Getter
		private List<IncomingFile> uploadedRepresentations;

		String create() {
			vorgangId = startCreation();

			uploadedAttachments = uploadAttachments();
			uploadedRepresentations = uploadRepresentations();

			finishCreation();
			return vorgangId;
		}

		String startCreation() {
			return vorgangRemoteStub.startCreation(buildStartCreationRequest(eingang)).getVorgangId();
		}

		private GrpcCreateVorgangRequest buildStartCreationRequest(GrpcEingang eingang) {
			var eingangWithoutFiles = eingang.toBuilder().clearAttachments().clearRepresentations().build();
			return GrpcCreateVorgangRequest.newBuilder().setEingang(eingangWithoutFiles).build();
		}

		List<IncomingFileGroup> uploadAttachments() {
			return formData.getAttachments().stream().map(this::uploadAttachment).toList();
		}

		private IncomingFileGroup uploadAttachment(IncomingFileGroup attachment) {
			var filesWithId = attachment.getFiles().stream().map(file -> file.toBuilder().id(uploadIncomingFile(file)).build()).toList();

			return IncomingFileGroup.builder().name(attachment.getName()).files(filesWithId).build();
		}

		List<IncomingFile> uploadRepresentations() {
			return formData.getRepresentations().stream()
					.map(representation -> representation.toBuilder().id(uploadIncomingFile(representation)).build()).toList();
		}

		String uploadIncomingFile(IncomingFile incomingFile) {
			var fileContentStream = incomingFile.getContentStreamForFinalRead();

			var resultFuture = GrpcFileUploadUtils.createSender(this::buildChunkRequest, fileContentStream,
					this::buildCallStreamObserver)
					.withMetaData(buildMetaDataRequest(incomingFile))
					.send();

			return waitUntilFutureToComplete(resultFuture, fileContentStream).getFileId();
		}

		GrpcUploadBinaryFileRequest buildChunkRequest(byte[] bytes) {
			return GrpcUploadBinaryFileRequest.newBuilder().setFileContent((ByteString.copyFrom(bytes))).build();
		}

		private CallStreamObserver<GrpcUploadBinaryFileRequest> buildCallStreamObserver(
				StreamObserver<GrpcUploadBinaryFileResponse> responseObserver) {
			return (CallStreamObserver<GrpcUploadBinaryFileRequest>) binaryFileRemoteStub.uploadBinaryFileAsStream(responseObserver);
		}

		GrpcUploadBinaryFileRequest buildMetaDataRequest(IncomingFile ingomingFile) {
			return GrpcUploadBinaryFileRequest.newBuilder()
					.setMetadata(GrpcUploadBinaryFileMetaData.newBuilder()
							.setContext(GrpcCallContext.newBuilder().setClient(CALL_CONTEXT_CLIENT).build())
							.setVorgangId(getVorgangId())
							.setField(VORGANG_ATTACHMENT_FIELD)
							.setContentType(ingomingFile.getContentType())
							.setSize(ingomingFile.getSize())
							.setFileName(ingomingFile.getName()))
					.build();
		}

		GrpcUploadBinaryFileResponse waitUntilFutureToComplete(CompletableFuture<GrpcUploadBinaryFileResponse> streamFuture,
				InputStream fileContentStream) {
			try {
				return streamFuture.get(10, TimeUnit.MINUTES);
			} catch (InterruptedException e) {
				Thread.currentThread().interrupt();
				throw new TechnicalException("Waiting for finishing upload was interrupted.", e);
			} catch (ExecutionException | TimeoutException e) {
				streamFuture.cancel(true);
				throw new TechnicalException("Error / Timeout on uploading data.", e);
			} finally {
				IOUtils.closeQuietly(fileContentStream);
			}
		}

		String finishCreation() {
			return vorgangRemoteStub.finishCreation(buildFinishCreationRequest()).getMessage();
		}

		GrpcFinishCreationRequest buildFinishCreationRequest() {
			return GrpcFinishCreationRequest.newBuilder()
					.addAllAttachments(getAttachmentsWithoutContent())
					.addAllRepresentations(getRepresentationsWithoutContent())
					.setVorgangId(getVorgangId())
					.build();
		}

		private List<GrpcIncomingFileGroup> getAttachmentsWithoutContent() {
			return getUploadedAttachments().stream().map(this::toIncomingFileGroup).toList();
		}

		private GrpcIncomingFileGroup toIncomingFileGroup(IncomingFileGroup incomingFileGroup) {
			return GrpcIncomingFileGroup.newBuilder()
					.setName(incomingFileGroup.getName())
					.addAllFiles(incomingFileGroup.getFiles().stream().map(this::toIncomingFile).toList()).build();
		}

		private List<GrpcIncomingFile> getRepresentationsWithoutContent() {
			return getUploadedRepresentations().stream().map(this::toIncomingFile).toList();
		}

		private GrpcIncomingFile toIncomingFile(IncomingFile incomingFile) {
			return GrpcIncomingFile.newBuilder().clearContent()
					.setId(incomingFile.getId())
					.setContentType(incomingFile.getContentType())
					.setName(incomingFile.getName())
					.setSize(incomingFile.getSize()).build();
		}
	}
}