/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.router;

import static org.assertj.core.api.Assertions.*;

import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;

import de.itvsh.ozg.pluto.vorgang.GrpcEingang;
import de.itvsh.ozg.pluto.vorgang.GrpcIncomingFile;
import de.ozgcloud.eingang.common.formdata.AntragstellerTestFactory;
import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.common.formdata.ZustaendigeStelleTestFactory;

@SpringBootTest
class GrpcEingangMapperITCase {

	@Autowired
	private GrpcEingangMapper grpcEingangMapper;

	@DisplayName("Mapped GrpcEingang")
	@Nested
	class TestToEingang {

		@Nested
		@DisplayName("Mapped Antragsteller")
		class TestMappingAntragsteller {

			@Test
			void antragstellerShouldBeMapped() {

				var antragSteller = grpcEingangMapper.toEingang(FormDataTestFactory.create()).getAntragsteller();

				assertThat(antragSteller.getPostfachId()).isEqualTo(AntragstellerTestFactory.POSTFACH_ID);
				assertThat(antragSteller.getVorname()).isEqualTo(AntragstellerTestFactory.VORNAME);
				assertThat(antragSteller.getTelefon()).isEqualTo(AntragstellerTestFactory.TELEFON);
			}

			@Test
			void dataShouldBeMapped() {

				var antragsteller = grpcEingangMapper.toEingang(FormDataTestFactory.create()).getAntragsteller();

				assertThat(antragsteller.getOtherData().getFieldList()).hasSize(1);
				assertThat(antragsteller.getOtherData().getField(0).getName()).isEqualTo(AntragstellerTestFactory.GEBIET_BEZEICHNUNG_KEY);
				assertThat(antragsteller.getOtherData().getField(0).getValue()).isEqualTo(AntragstellerTestFactory.GEBIET_BEZEICHNUNG_VALUE);
			}
		}

		@Nested
		@DisplayName("Mapped Zustaendinge Stelle")
		class TestZustaendigeStelle {
			@Test
			void eingangShouldHaveZustaendigeStelle() {

				var zustaendigeStelle = grpcEingangMapper.toEingang(FormDataTestFactory.create()).getZustaendigeStelle();

				assertThat(zustaendigeStelle).isNotNull();
				assertThat(zustaendigeStelle.getOrganisationseinheitenId()).isEqualTo(ZustaendigeStelleTestFactory.ORGANISATIONSEINHEIT_ID);
				assertThat(zustaendigeStelle.getEmail()).isEqualTo(ZustaendigeStelleTestFactory.EMAIL);
			}
		}

		@Nested
		@DisplayName("Test mapped Attachments")
		class TestAttachments {

			private GrpcEingang eingang;

			@BeforeEach
			void init() {

				eingang = grpcEingangMapper.toEingang(FormDataTestFactory.create());
			}

			@Test
			void validateNumberOfAttachments() {

				assertThat(eingang.getNumberOfAttachments()).isEqualTo(2);
			}

			@Test
			void validateNumberOfAttachmentGroups() {

				assertThat(eingang.getAttachmentsCount()).isEqualTo(2);
			}

			@Test
			void validateGroup1AttachmentCount() {

				assertThat(eingang.getAttachmentsList().get(0).getFilesCount()).isEqualTo(1);
			}

			@Test
			void validateGroup1Attachment() {

				GrpcIncomingFile attachment = eingang.getAttachmentsList().get(0).getFilesList().get(0);

				assertThat(attachment.getId()).isEqualTo(IncomingFileTestFactory.ID);
				assertThat(attachment.getVendorId()).isEqualTo(IncomingFileTestFactory.VENDOR_ID);
				assertThat(attachment.getName()).isEqualTo(IncomingFileTestFactory.NAME);
				assertThat(attachment.getContentType()).isEqualTo(IncomingFileTestFactory.CONTENT_TYPE);
				assertThat(attachment.getContent()).isEmpty();
			}

			@Test
			void validateGroup2AttachmentCount() {

				assertThat(eingang.getAttachmentsList().get(1).getFilesCount()).isEqualTo(1);
			}

			@Test
			void validateGroup2Attachment() {

				GrpcIncomingFile attachment = eingang.getAttachmentsList().get(1).getFilesList().get(0);

				assertThat(attachment.getId()).isEqualTo(IncomingFileTestFactory.ID);
				assertThat(attachment.getVendorId()).isEqualTo(IncomingFileTestFactory.VENDOR_ID);
				assertThat(attachment.getName()).isEqualTo(IncomingFileTestFactory.NAME);
				assertThat(attachment.getContentType()).isEqualTo(IncomingFileTestFactory.CONTENT_TYPE);
				assertThat(attachment.getContent()).isEmpty();
			}
		}

		@Nested
		@DisplayName("Test mapped Representations")
		class TestRepresentations {

			@Test
			void testRepresentations() {

				GrpcEingang eingang = grpcEingangMapper.toEingang(FormDataTestFactory.create());

				assertThat(eingang.getRepresentationsCount()).isEqualTo(1);

				GrpcIncomingFile representation = eingang.getRepresentationsList().get(0);
				assertThat(representation.getId()).isEqualTo(IncomingFileTestFactory.ID);
				assertThat(representation.getVendorId()).isEqualTo(IncomingFileTestFactory.VENDOR_ID);
				assertThat(representation.getName()).isEqualTo(IncomingFileTestFactory.NAME);
				assertThat(representation.getContentType()).isEqualTo(IncomingFileTestFactory.CONTENT_TYPE);
				assertThat(representation.getContent()).isEmpty();
			}
		}

		@Nested
		@DisplayName("Test mapping for formdata duplicate values")
		class TestDuplicateFormData {

			@Test
			void valueListShouldGenerateFields() {

				GrpcEingang eingang = grpcEingangMapper
						.toEingang(FormDataTestFactory.createBuilder().formData(Map.of("key", List.of("value1", "value2"))).build());

				assertThat(eingang.getFormData().getFieldCount()).isEqualTo(2);
			}

			@Test
			void objectListShouldGenerateSubForms() {

				GrpcEingang eingang = grpcEingangMapper
						.toEingang(FormDataTestFactory.createBuilder()
								.formData(Map.of("key-1", List.of(Map.of("sub_key", "value1"), Map.of("sub_key", "value2")))).build());

				assertThat(eingang.getFormData().getFormCount()).isEqualTo(2);
				assertThat(eingang.getFormData().getForm(0).getFieldCount()).isEqualTo(1);
			}
		}
	}
}
